/*
 * swfs_cluster.c
 *
 * Copyright (C) 2000, 2001 Richard Guenther
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <sys/types.h>
#include <sys/stat.h>
#include <pthread.h>
#include <unistd.h>
#include <fcntl.h>
#include "util.h"
#include "hash.h"
#include "pmap.h"
#include "swfs_cluster.h"


/* The in-memory cluster hash/lru and a lock protecting it. */
HASH(swcluster, struct swcluster, 10,
     (swcluster->name == name),
     (name),
     (swcluster->name),
     long name)
static struct list_head clusterslru;
static int clusterslrucnt;
static int clustersmaxlru;
static pthread_mutex_t clustersmx;
#define LOCKCLUSTERS do { pthread_mutex_lock(&clustersmx); } while (0)
#define UNLOCKCLUSTERS do { pthread_mutex_unlock(&clustersmx); } while (0)

/* Per cluster lock that protects the files list/cnt and the
 * cluster size/flags. */
#define LOCKCLUSTER(c) do {} while (0)
#define UNLOCKCLUSTER(c) do {} while (0)


/* The internal helpers (with no locking) */
static int __cluster_data_open(long name, int flags);
static int __cluster_meta_open(long name, int flags);
static struct swcluster *_cluster_creat(long name);
static struct swcluster *_cluster_stat(long name, s32 known_size);
static int _cluster_readfiles(struct swcluster *c);
static void _cluster_writefiles(struct swcluster *c);
static void _cluster_truncate(struct swcluster *c, s32 size);
static void _cluster_put(struct swcluster *c);
#define _cluster_needfiles(c) do { if ((c)->flags & SWC_NOT_IN_CORE) if (_cluster_readfiles(c) == -1) SWPANIC("metadata vanished under us"); } while (0)
#define _cluster_needdata(c) do { if ((c)->flags & SWC_CREAT || (c)->fd == -1) __cluster_needdata(c); } while (0)
static void __cluster_needdata(struct swcluster *c);


/***********************************************************************
 * API stuff. For documentation see headerfile.
 */

static int cluster_init(int maxlru, size_t pmapminfree)
{
	/* Initialize pmap subsystem. */
	if (pmap_init(pmapminfree) == -1)
		return -1;

	/* Init the lru. */
	clustersmaxlru = maxlru;
	clusterslrucnt = 0;
	INIT_LIST_HEAD(&clusterslru);

	/* The global cluster mx protecting hash & lru. */
	pthread_mutex_init(&clustersmx, NULL);
	srand(getpid()); /* hehe... */

	return 0;
}

static void cluster_cleanup()
{
	struct swcluster *c;

	/* Clear the cluster lru. */
	LOCKCLUSTERS;
	while ((c = list_gethead(&clusterslru, struct swcluster, lru))) {
		hash_remove_swcluster(c);
		list_del(&c->lru);
		_cluster_put(c);
	}
	clusterslrucnt = 0;
	UNLOCKCLUSTERS;

	/* Cleanup pmap subsystem. */
	pmap_close();

	/* Cleanup. */
	pthread_mutex_destroy(&clustersmx);
}

static struct swcluster *cluster_get(long name, int flags, s32 known_size)
{
	struct swcluster *c;

	LOCKCLUSTERS;
	if (!(c = hash_find_swcluster(name)))
		c = _cluster_stat(name, known_size);
	else {
		if (c->usage++ == 0) {
			list_del_init(&c->lru);
			clusterslrucnt--;
		}
	}
	UNLOCKCLUSTERS;
	if (!c)
		return NULL;
#ifdef DEBUG
	if (known_size != -1 && c->size != known_size)
		SWPANIC("User and cluster disagree about its size");
#endif

	/* Read the files list, if required. */
	if ((flags & CLUSTERGET_READFILES)
	    && (c->flags & SWC_NOT_IN_CORE)) {
		LOCKCLUSTER(c);
		if (c->flags & SWC_NOT_IN_CORE)
			if (_cluster_readfiles(c) == -1) {
				if (c->usage > 1)
					SWPANIC("metadata vanished under us");
				UNLOCKCLUSTER(c);
				cluster_put(c, CLUSTERPUT_FREE);
				return NULL;
			}
		UNLOCKCLUSTER(c);
	}

	return c;
}

static void cluster_put(struct swcluster *c, int flags)
{
	if (!c)
		return;

	/* Release the reference and do resouece cleanup, if
	 * this was the last one. */
	LOCKCLUSTERS;
	if (--(c->usage) == 0) {
		/* Free the cluster, if requested. */
		if (flags & CLUSTERPUT_FREE) {
			hash_remove_swcluster(c);
			UNLOCKCLUSTERS;
			_cluster_put(c);
			return;
		}

		/* Put cluster into the lru. */
		list_add(&c->lru, &clusterslru);
		clusterslrucnt++;
		if (clusterslrucnt > clustersmaxlru) {
			struct swcluster *c2;
			c2 = list_gettail(&clusterslru, struct swcluster, lru);
			hash_remove_swcluster(c2);
			list_del(&c2->lru);
			clusterslrucnt--;
			UNLOCKCLUSTERS;
			_cluster_put(c2);
			LOCKCLUSTERS;
		}
	}
	UNLOCKCLUSTERS;

	/* Check, if we are instructed to sync the metadata. */
	if ((flags & CLUSTERPUT_SYNC) && (c->flags & SWC_DIRTY)) {
		LOCKCLUSTER(c);
		if (c->flags & SWC_DIRTY)
			_cluster_writefiles(c);
		UNLOCKCLUSTER(c);
	}
}

static struct swcluster *cluster_alloc(s32 size)
{
	static pthread_mutex_t mx = PTHREAD_MUTEX_INITIALIZER;
	struct swcluster *c;
	long cid;

	/* This lock protects against racing free cid finding. */
	pthread_mutex_lock(&mx);

	/* Try, until we find an unallocated cid. This method is
	 * good if the cluster space is very sparse allocated and
	 * gets nearly unusable if more than the half of all clusters
	 * are allocated... */
	cid = rand();
	while ((c = cluster_get(cid, 0, -1))) {
		cluster_put(c, 0);
		cid = rand();
	}

	if ((c = _cluster_creat(cid)))
		c->size = size;

	pthread_mutex_unlock(&mx);

	return c;
}

static void cluster_addfileref(struct swcluster *c, long file)
{
	LOCKCLUSTER(c);
	/* Bring in the files list, if necessary. */
	_cluster_needfiles(c);

	/* Re-alloc the files array. - FIXME: inefficient */
	if (!(c->files = (long *)realloc(c->files, (c->files_cnt+1)*sizeof(long))))
		SWPANIC("cannot realloc files array");

	/* Append the file, fix the files count and mark it dirty. */
	c->files[c->files_cnt] = file;
	c->files_cnt++;
	c->flags |= SWC_DIRTY;
	UNLOCKCLUSTER(c);
}

static int cluster_delfileref(struct swcluster *c, long file)
{
	int i;

	LOCKCLUSTER(c);
	/* Bring in the files list, if necessary. */
	_cluster_needfiles(c);

	/* Search file from the back (this is most cache friendly
	 * for the necessary array move afterwards). */
	for (i=c->files_cnt-1; i>=0; i--) {
		if (c->files[i] == file) {
			/* Fix the files count. */
			c->files_cnt--;

			/* Start moving every file after i one
			 * position to the front of the list. */
			for (; i<c->files_cnt; i++)
				c->files[i] = c->files[i+1];

			c->flags |= SWC_DIRTY;
			UNLOCKCLUSTER(c);
			return 0;
		}
	}

	/* Not found!? */
	UNLOCKCLUSTER(c);
	return -1;
}

static int cluster_checkfileref(struct swcluster *c, long file)
{
	int i;

	LOCKCLUSTER(c);
	/* Bring in the files list, if necessary. */
	_cluster_needfiles(c);

	/* Search file from the back (this is most cache friendly
	 * for the necessary array move afterwards). */
	for (i=c->files_cnt-1; i>=0; i--) {
		if (c->files[i] == file) {
			UNLOCKCLUSTER(c);
			return 0;
		}
	}

	/* Not found!? */
	UNLOCKCLUSTER(c);
	return -1;
}

static char *_cluster_mmap(struct swcluster *c,void *start,
			   int prot, int flags)
{
	_cluster_needdata(c);
	return (char *)pmap_map(start, c->size, prot, flags, c->fd, 0);
}

static char *cluster_mmap(struct swcluster *c,void *start, int prot, int flags)
{
	char *mem;

	LOCKCLUSTER(c);
	_cluster_needdata(c);
	mem = (char *)pmap_map(start, c->size, prot, flags, c->fd, 0);
	UNLOCKCLUSTER(c);

	return mem;
}

static int cluster_munmap(char *start)
{
	if (pmap_unmap(start) == -1) {
		DPRINTF("Unable to unmap %p\n", start);
		return -1;
	} else
		return 0;
}


static void cluster_split(struct swcluster *c, s32 offset, s32 cutcnt,
			  struct swcluster **ch, struct swcluster **ct)
{
	char *m, *mh, *mt;

	if (!c || offset < 0 || offset > c->size
	    || cutcnt < 0 || offset+cutcnt > c->size
	    || (!ch && !ct))
		DERROR("Illegal use of cluster_split");

	/* FIXME wrt mappings! */
	if (c->fd != -1
	    && pmap_uncache(c->size, PROT_READ|PROT_WRITE,
			    MAP_SHARED|MAP_PRIVATE, c->fd, 0) == -1) {
		DPRINTF("Warning! Possibly now corrupt mappings!\n");
		pmap_invalidate(c->size, PROT_READ|PROT_WRITE,
				MAP_SHARED|MAP_PRIVATE, c->fd, 0);
	}

	LOCKCLUSTER(c);

	if (c->files_cnt == 1) {
		/* If the cluster is not shared, things are a lot
		 * simpler. */
		if (ct && !ch) {
			if ((m = _cluster_mmap(c, NULL, PROT_READ|PROT_WRITE, MAP_SHARED)) == MAP_FAILED)
				SWPANIC("Cannot mmap cluster");
			memmove(m, m + offset + cutcnt,
				c->size - offset - cutcnt);
			cluster_munmap(m);
			_cluster_truncate(c, c->size - offset - cutcnt);
			*ct = cluster_get(c->name, 0, c->size);
		} else if (!ct && ch) {
			_cluster_truncate(c, offset);
			*ch = cluster_get(c->name, 0, c->size);
		} else /* if (ct && ch) */ {
			if (!(*ct = cluster_alloc(c->size - offset - cutcnt))
			    || (m = _cluster_mmap(c, NULL, PROT_READ,
						  MAP_SHARED)) == MAP_FAILED
			    || (mt = _cluster_mmap(*ct, NULL, PROT_WRITE,
						   MAP_SHARED)) == MAP_FAILED)
				SWPANIC("Cannot alloc/mmap cluster");
			memcpy(mt, m + offset + cutcnt,
			       c->size - offset - cutcnt);
			cluster_munmap(mt);
			cluster_munmap(m);
			_cluster_truncate(c, offset);
			*ch = cluster_get(c->name, 0, c->size);
		}
	} else {
		/* Umm, now we have to copy in any case. */
		if ((m = _cluster_mmap(c, NULL, PROT_READ,
				       MAP_SHARED)) == MAP_FAILED)
			SWPANIC("Cannot mmap cluster");
		if (ch) {
			if (!(*ch = cluster_alloc(offset))
			    || (mh = _cluster_mmap(*ch, NULL, PROT_WRITE,
						   MAP_SHARED)) == MAP_FAILED)
				SWPANIC("Cannot alloc/mmap cluster");
			memcpy(mh, m, offset);
			cluster_munmap(mh);
		}
		if (ct) {
			if (!(*ct = cluster_alloc(c->size - offset - cutcnt))
			    || (mt = _cluster_mmap(*ct, NULL, PROT_WRITE,
						   MAP_SHARED)) == MAP_FAILED)
				SWPANIC("Cannot alloc/mmap cluster");
			memcpy(mt, m + offset + cutcnt,
			       c->size - offset - cutcnt);
			cluster_munmap(mt);
		}
		cluster_munmap(m);
	}

	UNLOCKCLUSTER(c);
}

static int cluster_truncate(struct swcluster *c, s32 size)
{
	LOCKCLUSTER(c);

	if (c->files_cnt > 1
	    || (c->fd != -1
		&& pmap_uncache(c->size, PROT_READ|PROT_WRITE,
				MAP_SHARED|MAP_PRIVATE, c->fd, 0) == -1)) {
		UNLOCKCLUSTER(c);
		return -1;
	}
	_cluster_truncate(c, size);

	UNLOCKCLUSTER(c);

	return 0;
}

static struct swcluster *cluster_unshare(struct swcluster *c)
{
	struct swcluster *cc;
	char *m, *mc;

	LOCKCLUSTER(c);

	/* If we have only one reference to a file, we might just
	 * return the cluster. */
	_cluster_needfiles(c);
	if (c->files_cnt <= 1) {
		UNLOCKCLUSTER(c);
		return c;
	}

	/* Now we have to copy the whole cluster (and return the copy
	 * instead of the original). */
	if (!(cc = cluster_alloc(c->size))
	    || (m = _cluster_mmap(c, NULL, PROT_READ, MAP_SHARED)) == MAP_FAILED
	    || (mc = _cluster_mmap(cc, NULL, PROT_WRITE, MAP_SHARED)) == MAP_FAILED)
		SWPANIC("Cannot alloc/mmap clusters");
	memcpy(mc, m, c->size);
	UNLOCKCLUSTER(c);

	cluster_munmap(mc);
	cluster_munmap(m);

	return cc;
}


static ssize_t cluster_read(struct swcluster *c, void *buf,
			    size_t count, off_t offset)
{
	ssize_t res;

	LOCKCLUSTER(c);
	_cluster_needdata(c);
	res = pread(c->fd, buf, count, offset);
	UNLOCKCLUSTER(c);

	return res;
}

static ssize_t cluster_write(struct swcluster *c, const void *buf,
			     size_t count, off_t offset)

{
	ssize_t res;

	LOCKCLUSTER(c);
	if (c->size < offset+count)
		DERROR("Write extends cluster size");
	_cluster_needdata(c);
	res = pwrite(c->fd, buf, count, offset);
	UNLOCKCLUSTER(c);

	return res;
}

static ssize_t cluster_write_extend(struct swcluster *c, const void *buf,
				    size_t count)

{
	ssize_t res;

	LOCKCLUSTER(c);
	_cluster_needdata(c);
	res = pwrite(c->fd, buf, count, c->size);
	c->size += res;
	UNLOCKCLUSTER(c);

	return res;
}



/***********************************************************************
 * Internal helpers. No locking.
 */

/* Helpers for opening the clusters metadata or data file on disk. */
static int __cluster_meta_open(long name, int flags)
{
	char s[256];

	snprintf(s, 255, "%s/%li", swap.clusters_meta_base, name);
	return open(s, flags, 0666);
}
static int __cluster_data_open(long name, int flags)
{
	char s[256];

	snprintf(s, 255, "%s/%li", swap.clusters_data_base, name);
	return open(s, flags, 0666);
}

/* Create a new in-memory representation of a cluster with name
 * entirely from scratch. To be useful you have to know no on-disk
 * or in-memory representation for this cluster exists. Initial
 * state is SWC_DIRTY|SWC_CREAT.
 * Returns a cluster reference or NULL if no memory was available. */
static struct swcluster *_cluster_creat(long name)
{
	struct swcluster *c;

	/* Allocate cluster structure. */
	if (!(c = (struct swcluster *)malloc(sizeof(struct swcluster))))
		return NULL;
	hash_init_swcluster(c);
	INIT_LIST_HEAD(&c->lru);
	c->name = name;
	c->usage = 1;
	c->flags = SWC_DIRTY|SWC_CREAT;
	c->size = 0;
	c->fd = -1;
	c->files_cnt = 0;
	c->files = NULL;

	hash_add_swcluster(c);
	return c;
}

/* Create a new in-memory representation of a cluster. Information
 * is generated using the on-disk representation - and an optional
 * provided size of the cluster (this will skip the stat of the
 * on-disk representation), which can be -1 if you dont know it or
 * if the cluster exists at all.
 * Returns a cluster reference, or NULL if anything went wrong.
 * Initial state of the cluster will be SWC_NOT_IN_CORE. */
static struct swcluster *_cluster_stat(long name, s32 known_size)
{
	struct swcluster *c;
	struct stat dstat;
	char sd[256];

	/* Allocate cluster structure. */
	if (!(c = (struct swcluster *)malloc(sizeof(struct swcluster))))
		return NULL;

	/* Stat metadata and data. */
#ifndef DEBUG
	if (known_size == -1) {
#endif
		snprintf(sd, 255, "%s/%li", swap.clusters_data_base, name);
		if (stat(sd, &dstat) == -1) {
			free(c);
			return NULL;
		}
		known_size = dstat.st_size;
#ifndef DEBUG
	}
#endif

	hash_init_swcluster(c);
	INIT_LIST_HEAD(&c->lru);
	c->name = name;
	c->usage = 1;
	c->flags = SWC_NOT_IN_CORE;
	c->size = known_size;
	c->fd = -1;
	c->files_cnt = -1;
	c->files = NULL;

	hash_add_swcluster(c);
	return c;
}

/* Reads the list of cluster users into memory. It is an error
 * to call this operation on a SWC_DIRTY cluster. For clusters
 * which already have the list this is a nop. Any internal error
 * causes a SWPANIC. Clears SWC_NOT_IN_CORE flag. */
static int _cluster_readfiles(struct swcluster *c)
{
	int fd;
	struct stat stat;

	if (c->flags & SWC_DIRTY)
		DERROR("read into dirty state");
	if (!(c->flags & SWC_NOT_IN_CORE))
		return 0;

	/* Load files list. */
	if ((fd = __cluster_meta_open(c->name, O_RDWR)) == -1
	    || fstat(fd, &stat) == -1) {
		if (fd != -1)
			close(fd);
		return -1;
	}
	if (!(c->files = (long *)malloc(stat.st_size)))
		SWPANIC("no memory for files list");
	c->files_cnt = stat.st_size/sizeof(long);
	if (read(fd, c->files, stat.st_size) != stat.st_size)
		SWPANIC("cannot read cluster metadata");
	c->flags &= ~SWC_NOT_IN_CORE;
	close(fd);

	return 0;
}

/* Writes the list of cluster users - the file list - to
 * the on-disk metadata file of the cluster. This file gets
 * created, if it is necessary.
 * If the cluster is not SWC_DIRTY this has no effect. The
 * SWC_DIRTY flag is cleared. Also for SWC_NOT_IN_CORE this
 * is a nop, too.
 * Failure is not permitted at any stage of this operation,
 * so we SWPANIC in this case. */
static void _cluster_writefiles(struct swcluster *c)
{
	int fd, size;

	if (!(c->flags & SWC_DIRTY)
	    || c->flags & SWC_NOT_IN_CORE)
		return;

	/* Open the metadatafile for writing (creating it, if it
	 * does not already exist), truncate it to the right
	 * size and write the file ids to the file. */
	size = c->files_cnt*sizeof(long);
	if ((fd = __cluster_meta_open(c->name, O_RDWR|O_CREAT)) == -1
	    || ftruncate(fd, size) == -1
	    || write(fd, c->files, size) != size)
		SWPANIC("cannot write cluster metadata");
	c->flags &= ~SWC_DIRTY;
	close(fd);
}

/* Create the clusters datafile, if necessary, and truncate it to
 * the right size. */
static void __cluster_needdata(struct swcluster *c)
{
	if (!(c->flags & SWC_CREAT)) {
		if (c->fd != -1)
			return;
		if ((c->fd = __cluster_data_open(c->name, O_RDWR)) == -1)
			SWPANIC("Cannot open cluster");
	} else {
		if (c->fd == -1
		    && (c->fd = __cluster_data_open(c->name, O_RDWR|O_CREAT)) == -1)
			SWPANIC("Cannot create cluster");
		if (ftruncate(c->fd, c->size) == -1)
			SWPANIC("Cannot truncate cluster");
		c->flags &= ~SWC_CREAT;
	}
}

/* Set the clusters size and truncate the on-disk file, if it
 * is already open, else just postpone that via SWC_CREAT. */
static void _cluster_truncate(struct swcluster *c, s32 size)
{
	/* If this is growing operation and the cluster
	 * had set SWC_CREAT (possibly not done previous
	 * shrinking truncate) we have to force that operation
	 * now to preserve correct zeroing semantics. */
	if (c->size < size && (c->flags & SWC_CREAT))
		_cluster_needdata(c);
	c->size = size;
	if (c->fd == -1)
		c->flags |= SWC_CREAT;
	else
		if (ftruncate(c->fd, c->size) == -1)
			SWPANIC("Cannot truncate cluster");
}

static void _cluster_put(struct swcluster *c)
{
	char s[256];

	/* Close the data fd - or create the file, if
	 * necessary. */
	if ((c->flags & SWC_CREAT) && c->files_cnt != 0)
		_cluster_needdata(c);
	/* Kill possible mappings and close the file. */
	if (c->fd != -1) {
		pmap_invalidate(c->size, PROT_READ|PROT_WRITE,
				MAP_SHARED|MAP_PRIVATE, c->fd, 0);
		close(c->fd);
	}

	/* If there are no users (files) left, unlink the
	 * on-disk representations. */
	if (!(c->flags & SWC_NOT_IN_CORE) && c->files_cnt == 0) {
		snprintf(s, 255, "%s/%li", swap.clusters_meta_base, c->name);
		unlink(s);
		snprintf(s, 255, "%s/%li", swap.clusters_data_base, c->name);
		unlink(s);
	} else if (c->flags & SWC_DIRTY)
		_cluster_writefiles(c);
	if (c->files)
		free(c->files);
	free(c);
}
